use std::path::PathBuf;

use elf_utilities::file::{ELF64Dumper, ELF64};

use crate::{
    common::{Output, Section},
    error::Error,
};

use super::segment::*;
use super::ElfObject;

pub struct ElfOutput<'data> {
    destination: PathBuf,
    file: ELF64,
    segment_data: SegmentData<'data>,
}

impl<'data> ElfOutput<'data> {
    pub fn new(destination: PathBuf) -> Self {
        use elf_utilities::header::{Class, Data, Machine, Type, Version, OSABI};

        let mut elf = ELF64::default();
        elf.ehdr.set_elf_type(Type::Exec);
        elf.ehdr.set_class(Class::Bit64);
        elf.ehdr.set_machine(Machine::Intel386);
        elf.ehdr.set_object_version(Version::Current);
        elf.ehdr.set_file_version(Version::Current);
        elf.ehdr.set_osabi(OSABI::Linux);
        elf.ehdr.set_data(Data::LSB2);

        Self {
            destination,
            file: elf,
            segment_data: SegmentData::new(),
        }
    }

    pub fn from_object(object: &ElfObject, destination: PathBuf) -> Self {
        let other = object.elf();

        Self {
            destination,
            file: elf_bin_from_object(other),
            segment_data: SegmentData::new(),
        }
    }

    fn populate_sections(&mut self) -> Result<u64, Error> {
        use elf_utilities::section::{Contents64, Section64, Shdr64};

        for (name, sections) in self.segment_data.iter_mut() {
            let mut data = Vec::new();

            for t in sections.iter_mut() {
                if let Some(iter) = &mut t.data {
                    data.extend(iter.as_mut());
                }
            }

            let section = Section64 {
                name: name.into(),
                header: Shdr64::default(),
                contents: Contents64::Raw(data),
            };
            self.file.add_section(section);
        }

        Ok(0)
    }

    fn populate_segment(&mut self, offset: &mut u64, size: u64) {
        use elf_utilities::segment::{Phdr64, Segment64, Type};
        use std::mem::size_of;

        let mut segment = Segment64 {
            header: Phdr64::default(),
        };
        segment.header.set_type(Type::Load);
        segment.header.p_filesz = size;
        segment.header.p_memsz = size;
        segment.header.p_offset = *offset;

        // TODO: elfix - add to elf-utilities as add_segment
        self.file.segments.push(segment);
        self.file.ehdr.e_phnum += 1;
        self.file.ehdr.e_phentsize = size_of::<Phdr64>() as u16;

        *offset += size;
    }

    fn populate_segments(&mut self) -> Result<u64, Error> {
        let mut offset = 0u64;
        // program header/segments
        // contains .text + .rodata as one segment
        self.populate_segment(&mut offset, self.segment_data.program_size());
        // contains .data as one segment
        self.populate_segment(&mut offset, self.segment_data.data_size());
        // contains .bss as one segment
        self.populate_segment(&mut offset, self.segment_data.bss_size());

        Ok(0)
    }
}

impl<'data> Output<'data> for ElfOutput<'data> {
    fn allocate(&mut self, size: u64) -> Result<u64, Error> {
        Ok(size)
    }

    fn append_section(&mut self, section: Section<'data>) -> Result<(), Error> {
        self.segment_data.append_section(section)
    }

    fn finalize(mut self) -> Result<PathBuf, Error> {
        use std::io::Error as IOError;
        use std::io::ErrorKind;

        self.populate_sections()?;
        self.populate_segments()?;
        self.file.finalize();

        let str_path = self.destination.to_str().ok_or_else(|| {
            let ioe = IOError::new(ErrorKind::Other, "Path expansion fail");
            let boxed = Box::new(ioe);
            Error::IOError(boxed)
        })?;

        ELF64Dumper::new(self.file)
            .generate_elf_file(str_path, 0o755)
            .map_err(|e| Error::IOError(e))?;

        Ok(self.destination)
    }
}

// init new ELF64 from an object file meant as executable output
fn elf_bin_from_object(other: &ELF64) -> ELF64 {
    use elf_utilities::header::Type;

    let mut elf = ELF64::default();
    elf.ehdr.set_elf_type(Type::Exec);
    elf.ehdr.e_ehsize = other.ehdr.e_ehsize;
    elf.ehdr.e_version = other.ehdr.e_version;
    elf.ehdr.e_ident = other.ehdr.e_ident;
    elf.ehdr.e_machine = other.ehdr.e_machine;

    elf
}